package industries.doublehelix.idleclock;

import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import com.github.kwhat.jnativehook.GlobalScreen;
import com.github.kwhat.jnativehook.NativeHookException;

import industries.doublehelix.idleclock.listeners.ChangeListener;
import industries.doublehelix.idleclock.listeners.ChangeListener.ActivityChangeReporter;

public class IdleClock {
	static DateTimeFormatter formatter = DateTimeFormatter.ofLocalizedDateTime(FormatStyle.LONG)
                     .withLocale(Locale.GERMANY)
                     .withZone(ZoneId.systemDefault());
	
	public static void main(String[] args) {
		// read settings
		Properties prop = getProperties();
		int idleTime = (int)TimeUnit.MINUTES.toSeconds(20);
		ReportType reportType = ReportType.IDLE;
		int linesToKeep = 5;
		Path reportFilePath = Paths.get(System.getProperty("user.home"), "idleClock.txt");
		
		String propIdleTime = prop.getProperty("idleTime", Integer.toString(idleTime));
		try {
			idleTime = Integer.parseInt(propIdleTime);
		} catch (NumberFormatException e) {
			System.err.println("idleTime \"" + propIdleTime + "\" is not a number, defaulting to " + idleTime);
		}
		
		ReportType propReportType = ReportType.getReportTypeFromString(prop.getProperty("reportType", reportType.name()));
		if (propReportType == null) {
			System.err.println("reportType \"" + prop.getProperty("reportType") + "\" not recognized, defaulting to " + reportType.name());
		} else {
			reportType = propReportType;
		}
		
		String propLinesToKeep = prop.getProperty("lines", Integer.toString(linesToKeep));
		try {
			linesToKeep = Integer.parseInt(propLinesToKeep);
		} catch (NumberFormatException e) {
			System.err.println("lines \"" + propLinesToKeep + "\" is not a number, defaulting to " + linesToKeep);
		}
		
		Path propReportFilePath = Paths.get(prop.getProperty("reportFile", reportFilePath.toString()));
		if (Files.isDirectory(propReportFilePath) || propReportFilePath.getParent() == null
				|| !Files.isDirectory(propReportFilePath.getParent())) {
			System.err.println("reportFile \"" + propReportFilePath.toString() + "\" is not a valid file, defaulting to " + reportFilePath.toString());
		} else {
			reportFilePath = propReportFilePath;
		}
		
		System.out.println("===CONFIGURATION LOADED===");
		System.out.println("Idling time: " + idleTime + " seconds");
		System.out.println("Reporting type: " + reportType.name());
		System.out.println("Lines to keep: " + linesToKeep);
		System.out.println("Report to file: " + reportFilePath.toString());
		
		
		
		// Clear previous logging configurations.
		LogManager.getLogManager().reset();

		// Get the logger for "org.jnativehook" and set the level to off.
		Logger logger = Logger.getLogger(GlobalScreen.class.getPackage().getName());
		logger.setLevel(Level.SEVERE);
		try {
			GlobalScreen.registerNativeHook();
		} catch (NativeHookException ex) {
			System.err.println("There was a problem registering the native hook.");
			System.err.println(ex.getMessage());
			System.exit(1);
		}
		
		final Path reportingFile = reportFilePath;
		final ReportType reportTypeFin = reportType;
		final int linesToKeepFin = linesToKeep;
		
		ChangeListener cl = new ChangeListener(10, idleTime);
		cl.addActivityChangeReporter(new ActivityChangeReporter() {
			@Override
			public void activityChanged(Instant lastActivityChangedInstant, ActivityChangeType type) {
				String idlingIdentifier = "Idling";
				String activityIdentifier = "Activity";
				
				String dateStr = formatter.format(lastActivityChangedInstant);
				String reportStr = "";
				if (reportTypeFin == ReportType.ACTIVE) {
					reportStr += type == ActivityChangeType.ACTIVITY_STARTED ? (activityIdentifier + " started: ") : "Stopped: ";
				} else {
					reportStr += type == ActivityChangeType.ACTIVITY_STARTED ? "Stopped: " : (idlingIdentifier + " started: ");
				}
				
				reportStr = reportStr + dateStr;
				System.out.println(reportStr);
				
				List<String> fileContents;
				// report to file
				// if file exists, prune to <linesToKeepFin> lines and report according to reporttype
				boolean fileExists = Files.exists(reportingFile);
				if (fileExists) {
					try {
						fileContents = Files.readAllLines(reportingFile, StandardCharsets.UTF_8);
					} catch (IOException e) {
						// act like file was empty
						e.printStackTrace();
						fileContents = new ArrayList<>();
					}
					int lastLineIdx = fileContents.size() - 1;
					String lastLine = lastLineIdx >= 0 ? fileContents.get(lastLineIdx) : null;
					
					// if last line has only first activity of current logging type and current activity is second activity,
					// add to line. If not, set line to null as if there was no line
					if (lastLine.contains("Stopped")) lastLine = null; // line contains second activity
					if (lastLine != null) {
						if (lastLine.startsWith(idlingIdentifier) && reportTypeFin == ReportType.IDLE
								&& type == ActivityChangeType.ACTIVITY_STARTED) {
							lastLine = lastLine + "\t" + reportStr;
						} else if (lastLine.startsWith(activityIdentifier) && reportTypeFin == ReportType.ACTIVE
								&& type == ActivityChangeType.ACTIVITY_STOPPED) {
							lastLine = lastLine + "\t" + reportStr;
						} else {
							lastLine = null;
						}
					}
					
					if (lastLine != null) {
						fileContents.set(lastLineIdx, lastLine);
					} else {
						if (type == ActivityChangeType.ACTIVITY_STOPPED && reportTypeFin == ReportType.ACTIVE) {
							fileContents.add(activityIdentifier + " started: Unknown\t" + reportStr);
						} else if (type == ActivityChangeType.ACTIVITY_STARTED && reportTypeFin == ReportType.IDLE) {
							// this can't happen, as the start of the program would count as the start, but eh, better be
							// save than sorry
							fileContents.add(idlingIdentifier + " started: Unknown\t" + reportStr);
						} else {
							fileContents.add(reportStr);
						}
					}
					
					if (fileContents.size() > linesToKeepFin) {
						fileContents = fileContents.subList(fileContents.size() - linesToKeepFin, fileContents.size());
					}
				} else {
					fileContents = new ArrayList<>();
					fileContents.add(reportStr);
				}
				
				// append to logging file if it exists, else create it
				try {
					Files.write(reportingFile, fileContents, StandardCharsets.UTF_8,
							fileExists ? StandardOpenOption.TRUNCATE_EXISTING : StandardOpenOption.CREATE);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		});
		cl.start();
	}
	
	public static Properties getProperties() {
		Path propertiesFile = null;
		// are we developing or running inside a jar?
		String resourcePath = IdleClock.class.getResource("IdleClock.class").toString();
		if (resourcePath.startsWith("jar:") || resourcePath.startsWith("rsrc:")){
			try {
				propertiesFile = Paths.get(IdleClock.class.getProtectionDomain().getCodeSource().getLocation().toURI());
				// aparently this was once the dir but on java 17 the jar itself
				if (Files.isRegularFile(propertiesFile)) {
					propertiesFile = propertiesFile.getParent();
				}
				propertiesFile = propertiesFile.resolve("IdleClock.properties");
				System.out.println("Evaluated properties path (next to jar): " + propertiesFile.toString());
			} catch (URISyntaxException e) {
				e.printStackTrace();
			}
		} else {
			propertiesFile = Paths.get(System.getProperty("user.dir"), "resources/IdleClock.properties"); // project dir in eclipse
			System.out.println("Evaluated properties path inside IDE: " + propertiesFile.toString());
		}
		if (propertiesFile == null || !Files.exists(propertiesFile)) {
			// see if one is in user home
			propertiesFile = Paths.get(System.getProperty("user.home"), "IdleClock.properties");
			System.out.println("Evaluated properties path inside user.home: " + propertiesFile.toString());
			if (!Files.exists(propertiesFile)) {
				System.err.println("Couldn't find IdleClock.properties, neither in jar dir nor user home, aborting");
				System.exit(1);
			}
		}
		Properties prop = new Properties();
		try(InputStream input = Files.newInputStream(propertiesFile)) {
			prop.load(input);
		} catch (Exception e1) {
			e1.printStackTrace();
			System.err.println("Could not load properties file, exiting");
			System.exit(1);
		}
		return prop;
	}
}
