package industries.doublehelix.idleclock.listeners;

import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;

import com.github.kwhat.jnativehook.GlobalScreen;
import com.github.kwhat.jnativehook.keyboard.NativeKeyEvent;
import com.github.kwhat.jnativehook.keyboard.NativeKeyListener;

import industries.doublehelix.idleclock.listeners.ChangeListener.ActivityChangeReporter.ActivityChangeType;

public class ChangeListener implements Runnable, NativeKeyListener {
	private static final int MOUSE_INTERVAL = 1000;
	
	private Instant lastKeyChange = Instant.now();
	
	private Instant lastCheckedInstant = null;
	private Instant lastReportedInstant = null;
	private long interval = 1000; // ms
	private int minSecondsForIdle = 60;
	private volatile boolean stop = false;
	private ArrayList<ActivityChangeReporter> reporterList = new ArrayList<>();
	
	private MouseChangeListener mcl = null;

	/**
	 * 
	 * @param interval interval to check in seconds. Will be reduced to half of minSecondsForIdle if it is not at least that
	 * @param minSecondsForIdle calls all ActivityChangeReporters after x seconds of idling for idle state
	 * (active state if we were idling is reported instantly)
	 */
	public ChangeListener(int interval, int minSecondsForIdle) {
		this.minSecondsForIdle = minSecondsForIdle;
		
		if (interval > minSecondsForIdle / 2) {
			this.interval = ((minSecondsForIdle / 2) * 1000);
		} else {
			this.interval = (interval * 1000);
		}
	}
	
	public void start() {
		GlobalScreen.addNativeKeyListener(this);
		new Thread(this, "ChangeListener").start();
	}
	
	public void stop() {
		this.stop = true;
	}
	
	public void addActivityChangeReporter(ActivityChangeReporter ir) {
		this.reporterList.add(ir);
	}
	
	public void removeActivityChangeReporter(ActivityChangeReporter ir) {
		this.reporterList.remove(ir);
	}
	
	@Override
	public void run() {
		// init
		this.mcl = new MouseChangeListener(MOUSE_INTERVAL);
		this.lastCheckedInstant = Instant.now();
		this.lastReportedInstant = Instant.now();
		
		this.mcl.start();
		
		while(true) {
			if (this.stop) break;
			
			// try/finally block for delay 
			try {
				Instant lastMouseInstant = this.mcl.getLastInstant();
				
				Instant instantToCheck = this.lastKeyChange.isAfter(lastMouseInstant) ? this.lastKeyChange : lastMouseInstant;
				
				// skip all the work if we reported that already (can only happen when idling, not returning
				if (this.lastReportedInstant.equals(instantToCheck)) continue; 
				
				if (instantToCheck.isAfter(this.lastCheckedInstant)) {
					// check if the last instant was the set time ago or more, if so we are resuming activity
					Duration d = Duration.between(this.lastCheckedInstant , instantToCheck);
					if (d.getSeconds() >= this.minSecondsForIdle) {
						this.lastReportedInstant = this.lastCheckedInstant = instantToCheck;
						for (ActivityChangeReporter ir : this.reporterList) {
							ir.activityChanged(this.lastReportedInstant, ActivityChangeType.ACTIVITY_STARTED);
						}
						continue;
					}
					
					// as resuming is handled, we don't need to keep the last instant for that
					this.lastCheckedInstant = instantToCheck;
				}
				
				Instant current = Instant.now();
				Duration d = Duration.between(this.lastCheckedInstant , current);
		
				if (d.getSeconds() >= this.minSecondsForIdle) {
					this.lastReportedInstant = this.lastCheckedInstant;
					for (ActivityChangeReporter ir : this.reporterList) {
						ir.activityChanged(this.lastReportedInstant, ActivityChangeType.ACTIVITY_STOPPED);
					}
				}
			} finally {
				try {
					Thread.sleep(this.interval);
				} catch (InterruptedException e) {
					e.printStackTrace();
					break;
				}
			}
		}
	}
	
	public interface ActivityChangeReporter {
		public enum ActivityChangeType{ ACTIVITY_STARTED, ACTIVITY_STOPPED}
		
		/**
		 * reports if activity changed
		 * @param lastActivityChangedInstant time instant when the last time was when activity changed to active or to idle
		 * @param type if activity resumed or stopped
		 */
		public void activityChanged(Instant lastActivityChangedInstant, ActivityChangeType type);
	}

	@Override
	public void nativeKeyPressed(NativeKeyEvent arg0) {
		this.lastKeyChange = Instant.now();
	}

	@Override
	public void nativeKeyReleased(NativeKeyEvent arg0) {
	}

	@Override
	public void nativeKeyTyped(NativeKeyEvent arg0) {
	}

}
